/* Bitbanging code to clock a JTAG sequence over the GPIO ports  of
 * the FT2232
 *
 */

#define JTAG_SWITCH_TMS 0
#define MAX_REG_SIZE    16
#define MAGIC_PAD_SIZE          4

// #include "jtag_intern.h"
// #include "jtag.h"
#include "emu.h"

const char *jtag_getstate(int state);
void hexdump(const unsigned char *buf, unsigned long n);

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#define PIN_TCK      0
#define PIN_TDI      1
#define PIN_TDO      2
#define PIN_TMS      3

#define GPIO_TCK  (1 << PIN_TCK)
#define GPIO_TDI  (1 << PIN_TDI)
#define GPIO_TDO  (1 << PIN_TDO)
#define GPIO_TMS  (1 << PIN_TMS)

// Those are outputs:
#define GPIO_DIR (GPIO_TCK | GPIO_TDI | GPIO_TMS)

unsigned char g_writebuf[0x10000];
unsigned char g_readbuf[0x10000];

const unsigned
char s_jtag_shift_ir[]          = { JTAG_SWITCH_TMS, 4-1, 0x03 };

const unsigned
char s_jtag_shift_dr[]          = { JTAG_SWITCH_TMS, 3-1, 0x01 };

const unsigned
char s_jtag_update[]            = { JTAG_SWITCH_TMS, 1-1, 0x03 };

const unsigned
char s_jtag_rti_frompause[]     = { JTAG_SWITCH_TMS, 3-1, 0x03 };

const unsigned
char s_jtag_shiftdr_frompause[] = { JTAG_SWITCH_TMS, 5-1, 0x07 };

const unsigned
char s_jtag_shiftir_frompause[] = { JTAG_SWITCH_TMS, 6-1, 0x0f };

const
char s_jtag_rtishiftdr_frompause[] = { JTAG_SWITCH_TMS, 5-1, 0x07 };


// in state EXIT1

const unsigned
char s_jtag_run_test_idle[]     = { JTAG_SWITCH_TMS, 1-1, 0x00 };

const unsigned
char s_jtag_pause[]             = { JTAG_SWITCH_TMS, 1-1, 0x00 };

const unsigned
char s_jtag_rti[]               = { JTAG_SWITCH_TMS, 2-1, 0x01 };

const unsigned
char s_jtag_test_logic_reset[]  = { JTAG_SWITCH_TMS, 5-1, 0x1f }; 

extern JTAG_CTRL g_jtag;

static void default_log(const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	vfprintf(stderr, fmt, args); fprintf(stderr, "\n");
	va_end(args);
}


void (*jtag_log)(const char *fmt, ...) = default_log;

int jtag_myread(Ctrl *c, unsigned char *buf, unsigned int len)
{
	int n;
	int size = 0;

	while (size < len) {
		n = ftdi_read_data(c->device, buf, len);
		if (n < 0) {
			printf("ftdi_read_data() error %d\n", n);
			return n;
		}
		size += n;
		len -= n;
		buf += n;
	}

	return 0;
}

int jtag_tms_sequence(Ctrl *c, const unsigned char *state)
{
	int error = 0;
	int i, n;
	unsigned char data, pins;
	static unsigned char seq[16];
	unsigned char *p;

	n = state[1] + 1;
	if (n > 8) {
		printf("too long TMS sequence\n");
		return -1;
	}
	data = state[2];

	p = seq;
	for (i = 0; i < n; i++) {
		pins = (data & 1) << PIN_TMS;
		*p++ = pins;
		*p++ = pins | GPIO_TCK;
		data >>= 1;
	}

	jtag_sendnow(c, seq, 2 * n);
	jtag_myread(c, g_readbuf, 2 * n);

	// error = jtag_myread(c, g_readbuf, 2 * n);

	return error;
}


int jtag_goto_state(Ctrl *c, const unsigned char *state)
{
	int error;
	int i, n;
	unsigned char data, pins;
	static unsigned char seq[16];
	unsigned char *p;

	n = state[1] + 1;
	data = state[2];

	p = seq;
	for (i = 0; i < n; i++) {
		pins = (data & 1) << PIN_TMS;
		c->state = g_state_links[c->state][data & 1];
		*p++ = pins;
		*p++ = pins | GPIO_TCK;
		data >>= 1;
	}

	jtag_sendnow(c, seq, 2 * n);
	error = jtag_myread(c, g_readbuf, 2 * n);

	// printf("     State: %s\n", jtag_getstate(c->state));

	return error;
}

int jtag_reset(Ctrl *c, int cycles)
{
	unsigned char seq[] = { 0, 5-1, 0x1f };
	jtag_goto_state(c, seq);
	return jtag_goto_state(c, s_jtag_run_test_idle);
}

#define FEED_TMS(c, tms, val, count)           \
	tms |= (val << count); count++;             \
	c->state = g_state_links[c->state][val & 1];

int jtag_moveto_state(Ctrl *c, int s)
{
	int count = 0;
	char tms = 0;
	static unsigned char seq[3];

	if (s == c->state) return 0;
		

	seq[0] = JTAG_SWITCH_TMS;

	if (s < c->state && s != RUN_TEST_IDLE) {
		// Clock a '1' until we reach the select state
		// (don't go into RTI)
		while (c->state != SELECT_DR_SCAN) {
			FEED_TMS(c, tms, 1, count);
		}
		// IR chain states
		if (s >= SELECT_IR_SCAN) {
			FEED_TMS(c, tms, 1, count);
		}

		seq[1] = count - 1; seq[2] = tms;
		jtag_tms_sequence(c, seq);
	}

	count = 0; tms = 0;

	while (s != c->state && count < 5) {
		if (g_state_links[c->state][0] == s) {
			FEED_TMS(c, tms, 0, count);
		} else {
			FEED_TMS(c, tms, 1, count);
		}
	}


	if (s == c->state) {
		if (!count) return 0;
		seq[1] = count - 1; seq[2] = tms;
		return jtag_tms_sequence(c, seq);
	}
	return ERR_STATE;
}


static
int shift_generic_msb(unsigned char *in, unsigned char *out,
		int nbits, int mode)
{

	printf("NOT SUPPORTED\n");
    
	// data: << [M******L] [M******L] .. [M**L....]
	// Bit:      ^N-1       ^N-9      ..   0

	return -1;
}

void parse_bitstream(unsigned char *out, unsigned char *in, int which, int len)
{
	unsigned char val;
	int i;

	in++;

	while (len > 8) {
		i = 8;
		val = 0;
		while (i--) {
			val >>= 1;
			val |= (((*in) >> which) << 7) & 0x80;
		}
		*out++ = val;
		len -= 8;
		in += 2;
	}
	val = 0;
	i = 8 - len;
	while (len--) {
		val >>= 1;
		val |= (((*in) >> which) << 7) & 0x80;
		in += 2;
	}
	// Align LSB to bit 0:
	val >>= i;
	*out = val;
}

int shift_generic_lsb(const unsigned char *in, unsigned char *out,
		int nbits, int mode)
{

	//
	// data: << [M******L] [M******L] [..M****L]
	// Bit:             0          8..   ^N-1

	int error;
	const unsigned char *p;
	unsigned char *s;
	unsigned char pins;
	int i;
	unsigned char val;
	p = in;

	while (nbits > 8) {
		i = 8;
		val = *p++;
		s = g_writebuf;
		while (i--) {
			pins = (val & 1) << PIN_TDI;
			*s++ = pins;
			*s++ = pins | GPIO_TCK;
			val >>= 1;
		}
		nbits -= 8;

		// printf("output: ");
		// hexdump(g_writebuf, 16);

		jtag_sendnow(g_jtag, g_writebuf, 16);
		// Read and parse:
		jtag_myread(g_jtag, g_readbuf, 16);

		// printf("input: ");
		// hexdump(g_readbuf, 16);

		if (out) {
			parse_bitstream(out, g_readbuf, PIN_TDO, 8);
			out++;
		}
	}
	val = *p;
	i = nbits;
	s = g_writebuf;
	while (nbits > 1) {
		pins = (val & 1) << PIN_TDI;
		*s++ = pins;
		*s++ = pins | GPIO_TCK;
		nbits--;
		val >>= 1;
	}

	// Last bit:
	switch (mode) {
		case STAY:
			pins = ((val & 1) << PIN_TDI);
			break;
		default:
			pins = GPIO_TMS | ((val & 1) << PIN_TDI);
		// Track state for last bit
			g_jtag->state = g_state_links[g_jtag->state][1];
			break;
	}

	*s++ = pins;
	*s++ = pins | GPIO_TCK;

	jtag_sendnow(g_jtag, g_writebuf, 2 * i);
	error = jtag_myread(g_jtag, g_readbuf, 2 * i);
	if (error < 0) return error;
	if (out)
		parse_bitstream(out, g_readbuf, PIN_TDO, i);
	
	switch (mode) {
		case UPDATE:
			error = jtag_goto_state(g_jtag, s_jtag_update); break;
		case EXEC:
			error = jtag_goto_state(g_jtag, s_jtag_rti); break;
	}

	return error;
}


int shift_generic(JtagRegister *in, JtagRegister *out, int nbits, int mode)
{
	int error;
	switch (in->flags & JTAGREG_SIGBIT_MASK) {
		case JTAGREG_MSB:
			error = shift_generic_msb(in->data, out ? out->data: 0,
					nbits, mode);
			break;
		case JTAGREG_LSB:
			error = shift_generic_lsb(in->data, out ? out->data : 0,
					nbits, mode);
			break;
		default:
			error = -1;
	}

	return error;

}


int jtag_config(Ctrl *c, int mode, int value)
{
	int error = ERR_UNSUPPORTED;
	switch (mode) {
		case JTAG_NDEVICES:
			error = ERR_NONE;
			c->ndevices = value;
			if (c->instructions) free(c->instructions);
			c->instructions =
				(REGISTER *) malloc(c->ndevices * sizeof(long));
			if (c->instructions == NULL) error = ERR_MALLOC;
			jtag_reset(c, 40);
			break;
		case JTAG_CLKSPEED:
			error = jtag_setclk(c, value);
			break;
		case JTAG_GPIO:
			error = jtag_cfg_gpio(c, (unsigned short) value);
			break;
	}
	return error;
}


int jtag_get(Ctrl *c, int token, void *value)
{
	switch (token) {
		case JTAG_NDEVICES:
			*( (int *) value) = c->ndevices;
			break;
		case JTAG_STATE:
			*( (int *) value) = c->state;
			break;
		case JTAG_IR:
			*( (JtagRegister **) value) = c->ir;
			break;
		case JTAG_IRLENGTH:
			if (c->ir) {
				*( (int *) value) = c->ir->nbits;
			} else {
				*( (int *) value) = 0;
			}
			break;
		default:
			return ERR_UNSUPPORTED;
	}
	return 0;
}

int detect_regsize(Ctrl *c, int max_size)
{
	int error;
	int retcode = ERR_DETECT;
	int l, i;
	JtagRegister *reg =
		reg_new(8 * (MAGIC_PAD_SIZE + max_size), JTAGREG_LSB);
	JtagRegister *out =
		reg_new(8 * (MAGIC_PAD_SIZE + max_size), JTAGREG_LSB);

	g_jtag = c;
	
	// create random pattern:

	srand(260872);

	for (i = 0; i < MAGIC_PAD_SIZE; i++) {
		reg->data[i] = rand();
	}

	// shift register of assumed IR length plus magic pattern
	error = shift_generic(reg, out, 8 * (MAGIC_PAD_SIZE + max_size), STAY);
	if (error < 0) {
		retcode = error;
		goto failure;
	}

#ifdef DEBUG
	printf("IN:  ");
	hexdump(reg->data, MAGIC_PAD_SIZE + max_size);
	printf("OUT: ");
	hexdump(out->data, MAGIC_PAD_SIZE + max_size);
#endif

	for (l = 1; l < max_size * 8; l++) {
		reg_shift(out, 1);

		if (memcmp(reg->data, out->data, MAGIC_PAD_SIZE) == 0) {
			retcode = l;
		}
	}
failure:
	reg_free(reg); reg_free(out);
	return retcode;
}

int jtag_detect(Ctrl *c)
{
	int error;
	int irlen;

	unsigned char seq[2];

	seq[0] = GPIO_TMS | GPIO_TCK;

	jtag_sendnow(c, seq, 1);
	error = jtag_myread(c, &seq[1], 1);
	if (error < 0) return error;

	jtag_reset(c, 40);

	jtag_goto_state(c, s_jtag_shift_ir);
	irlen = detect_regsize(c, MAX_REG_SIZE);
	DEB(printf("IR size: %d\n", irlen));

	if (irlen < 0) {
		return ERR_DETECT;
	}

	jtag_reset(c, 10);
	// jtag_goto_state(c, s_jtag_test_logic_reset);
	// jtag_goto_state(c, s_jtag_run_test_idle);

   	c->ir = reg_new(irlen, JTAGREG_LSB);

	reg_bitset(c->ir, 1); // BYPASS
	jtag_goto_state(c, s_jtag_shift_ir);
	error = shift_generic(c->ir, NULL, c->ir->nbits, UPDATE);

	if (error < 0) return error;

	jtag_goto_state(c, s_jtag_shift_dr);
	if (c->state != SHIFT_DR) {
		printf(" WRONG STATE !!!\n");
		return -1;
	}
	c->ndevices = detect_regsize(c, MAX_REG_SIZE);

	if (c->ndevices < 0)
		return ERR_DETECT;

	if (c->instructions == NULL)
		c->instructions = (REGISTER *) malloc(c->ndevices * sizeof(long));
	jtag_reset(c, 40);
	return c->ndevices;
}

int jtag_wait(Ctrl *c, int cycles)
{
	int error = 0;

	usleep(cycles);
	jtag_goto_state(c, s_jtag_pause);

	return error;
}	

int jtag_setclk(Ctrl *c, int value)
{
	return 0;
}
