/**\file xsvf.c
 *
 * XSVF player for ICEbear
 *
 * (c) 2006, Martin Strubel <hackfin@section5.ch>
 *
 * $Id: xsvf.c 671 2008-09-09 14:47:05Z strubi $
 *
 * See also Xilinx Appnote xapp058
 *
 */

#include "xsvf.h"
#include "jtag.h"
#include "jtag_intern.h"
#include "platform.h"
#include <string.h>
#include <stdio.h>

// calculate number of bytes from given number of
// bits
#define NBYTES(x)   ((x + 7) >> 3)

////////////////////////////////////////////////////////////////////////////
//
// PROTOS
//
int xsvf_play(JTAG_CTRL c);

extern const unsigned char s_jtag_rti_frompause[];
extern const unsigned char s_jtag_shift_frompause[];
extern const unsigned char s_jtag_shiftdr_frompause[];
extern const unsigned char s_jtag_shiftir_frompause[];
extern const unsigned char s_jtag_rti[];
extern const unsigned char s_jtag_rtishiftdr_frompause[];

////////////////////////////////////////////////////////////////////////////
// GLOBALS

char *g_statestrings[] = {
	"Test Logic/Reset",
	"Run Test/Idle",
	"Select DR Scan",
	"Capture DR",
	"Shift DR",
	"Exit1 DR",
	"Pause DR",
	"Exit2 DR",
	"Update DR",
	"Select IR Scan",
	"Capture IR",
	"Shift IR",
	"Exit1 IR",
	"Pause IR",
	"Exit2 IR",
	"Update IR"
};

JTAG_CTRL g_jtag;


////////////////////////////////////////////////////////////////////////////
// AUXILIARIES (DEBUG)

void hexdump(const unsigned char *buf, unsigned long n)
{
	int i = 0;
	int c = 0;

	while (i < n) {
		printf("%02x ", buf[i]);
		c++;
		if (c == 16) { c = 0; printf("\n"); }
		i++;
	}
	if (c)
		printf("\n");
}

////////////////////////////////////////////////////////////////////////////
// { TEST MODE
#ifdef TEST

#include <stdio.h>

#ifdef DEBUG
#define PRINTF(x, ...) printf(x, ##__VA_ARGS__)
#else
#define PRINTF(x, ...)
#endif

void reg_dump(JtagRegister *r)
{
	int i;
	for (i = 1; i <= r->nbits; i++) {
		if (reg_get(r, r->nbits - i, 1) & 1)
			printf("1");
		else
			printf("0");
	}
	printf("\n");
}

int mask_cmp(unsigned char *d1, unsigned char *d2, unsigned char *mask, int n)
{
	unsigned char m;

	while (n--) {
		m = *mask++;
		if ((*d1 & m) != (*d2 & m)) {
			return 1;
		}
		d1++; d2++;
	}
	return 0;
}
 
void jtag_show_state()
{
	int state;
	jtag_get(g_jtag, JTAG_STATE, &state);
	printf("State: %s\n", 
		g_statestrings[state]);
}
	

FILE *g_file;

int nextByte(void)
{
	return fgetc(g_file);
}

int nextBlock(unsigned char *buf, int size)
{
	int m;

	printf("reading %d bytes...\n", size);
   	m = fread(buf, 1, size, g_file);
	if (m != size) {
		printf("Error:\n");
		perror("reading file");
		printf("Read %d bytes instead of %d\n", m, size);
		exit(-1);
	}
	return size;
}

void skip_bytes(unsigned long size)
{
	while (size--) nextByte();
}

int main(int argc, char **argv)
{
	int error;
	int ndevs;

	char *jtag_argv[] = { "0:1", NULL };

	JTAG_CTRL jtag;

	if (argc == 2) {
		g_file = fopen(argv[1], "rb");
	} else {
		printf("Filename expected\n");
		return -1;
	}

	error = jtag_open(jtag_argv, &jtag);
	if (error < 0) {
		printf("Could not open ICEbear. Is it connected?\n"
		       "Try replugging it.\n");
		return -1;
	}

	ndevs = jtag_detect(jtag);
	if (ndevs < 0) {
		printf("Could not detect scan chain on target\n"
		       "Is target powered and connected to ICEbear?\n");
		jtag_close(jtag);
		return -1;
	}

	jtag_config(jtag, JTAG_CLKSPEED, 5);
	
	printf("Number of devices in JTAG chain: %d\n", ndevs);

	error = xsvf_play(jtag);
	if (error < 0) {
		PRINTF("Error: %d\n", error);
	}

	jtag_close(jtag);
	
	fclose(g_file);
	return 0;
}


void notify(char *txt)
{
	PRINTF(": %s", txt);
}

#define BLOCKSIZE 1024
unsigned char g_buffer[BLOCKSIZE];
unsigned char g_mask[BLOCKSIZE];

#else // TEST


#define PRINTF(x, ...)

unsigned char *g_buffer = g_ram.isp.buffer;

static
unsigned long g_offset = 0;

#include "uart.h"

#ifdef DEBUG

#define SEND_CR() uart_puts("\r\n")

void notify(char *txt)
{
	uart_puts(txt);
}

#else

#define notify(x)
#define SEND_CR()

#endif
#endif


//}/////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// AUXILIARIES

#define read_int32() read_val(32)

/** Read n byte value (integer) from XSVF stream
 */

long read_val(nbits)
{
	unsigned long val = 0;
	int n = NBYTES(nbits);
	int i;

	nextBlock(g_buffer, n);

	for (i = 0; i < n; i++) {
		val <<= 8;
		val |= g_buffer[i];
	}

	if (nbits < 32)
		val &= ( (1 << nbits) - 1);
	return val;
}

void reverse_buffer(unsigned char *buf, int n)
{
	int c;
	unsigned char *p, *q;

	if (n < 1) {
		printf("Illegal buffer size\n");
		return;
	}

	p = buf;
	q = &buf[n - 1];

	while (p < q) {
		c = *q; *q = *p; *p = c;
		p++; q--;
	}
}

const char *debug_instr(unsigned short instr)
{
	switch (instr) {
		case 0xff: return "BYPASS";
		case 0x03: return "SAMPLE";
		case 0x00: return "EXTEST";
		case 0x01: return "IDCODE";
		case 0xfd: return "USERCODE";
		case 0xfc: return "HIGHZ";
		case 0xe9: return "ISC_ENABLE_CLAMP";
		case 0xe4: return "ISC_ENABLEOTF";
		case 0xe8: return "ISC_ENABLE";
		case 0xe7: return "ISC_SRAM_READ";
		case 0xe6: return "ISC_SRAM_WRITE";
		case 0xed: return "ISC_ERASE";
		case 0xea: return "ISC_PROGRAM";
		case 0xee: return "ISC_READ";
		case 0xf0: return "ISC_INIT";
		case 0xc0: return "ISC_DISABLE";
		case 0x11: return "TEST_ENABLE";
		case 0x12: return "BULKPROG";
		case 0x14: return "ERASE_ALL";
		case 0x13: return "MVERIFY";
		case 0x15: return "TEST_DISABLE";
		case 0xe0: return "ISC_NOOP";
	}
	return "Unknown";
}

////////////////////////////////////////////////////////////////////////////
// XSVF handlers

/**
 * Shift data into instruction register
 */

int irscan(short nbits, short mode, unsigned long time)
{
	unsigned short instr;
	instr = read_val(nbits);
	JtagRegister *ir;
	int error;
	int state;


	if (mode & F_FROM_PAUSE_IR) {
		printf("From pause\n");
		jtag_goto_state(g_jtag, s_jtag_shiftir_frompause);
	} else {
		jtag_goto_state(g_jtag, s_jtag_shift_ir);
	}
	PRINTF("    instr: %x : %s\n", instr, debug_instr(instr));
	jtag_get(g_jtag, JTAG_STATE, &state);
	if (state != SHIFT_IR) {
		printf("ERROR: Not in Shift IR state (%s)\n",
				g_statestrings[state]);
		return -1;
	}

	jtag_get(g_jtag, JTAG_IR, &ir);
	reg_set(ir, 0, nbits, instr);

	error = shift_generic(ir, NULL, nbits, EXIT);
	if (error < 0) return -1;

	// Goto XENDIR state:
	if (mode & F_PAUSE_IR && time == 0) {
		jtag_goto_state(g_jtag, s_jtag_pause);
	} else {
		jtag_goto_state(g_jtag, s_jtag_rti);
		PRINTF("     Sleep: %ld us\n", time);
		usleep(time);
	}

	return 0;
}

/**
 * Unsupported: Set SDRMASK
 */

int set_sdrmasks(unsigned long nbits)
{
	int n = NBYTES(nbits);

	skip_bytes(n);
	PRINTF(":");
	skip_bytes(n);
	printf("unsupported\n");
	return -1;
}

void memory_dump(unsigned char *buf, unsigned long n)
{
	int i = 0;
	int c = 0;

	while (i < n) {
		printf("%02x ", buf[i]);
		c++;
		if (c == 16) { c = 0; printf("\n"); }
		i++;
	}
	if (c)
		printf("\n");
}

/**
 * Shift data into data register
 *
 * \param nbits  number of bits to be shifted in
 * \mode         mode: one of { F_BEGIN, F_CONTINUE, F_END }; see xsvf.h
 */
#if 0

int drscan(unsigned long nbits, short mode)
{
	int n = NBYTES(nbits);
	int error;
	int state;

	// Enter shift DR state:
	jtag_get(g_jtag, JTAG_STATE, &state);
	if (mode & F_BEGIN) {
		if (state == PAUSE_IR || state == PAUSE_DR) {
			jtag_goto_state(g_jtag, s_jtag_update);
		}
		jtag_goto_state(g_jtag, s_jtag_shift_dr);
	}

	jtag_get(g_jtag, JTAG_STATE, &state);
	if (state != SHIFT_DR) {
		printf("ERROR: Not in Shift DR state (%s)\n",
				g_statestrings[state]);
		return -1;
	}
	nextBlock(g_buffer, n);
	reverse_buffer(g_buffer, n);
	n--;


	printf("clock tdi out\n");
	memory_dump(g_buffer, n);
	error = jtag_bytes_tdiout_stay(g_buffer, n);
	if (error < 0) { printf("Error: %s\n", __FUNCTION__); return error; }

	// Shift out last remaining bits:
	nbits -= (n << 3);
	printf("nbits: %d\n", nbits);

	jtag_show_state();

	if (mode & F_END) {
		memory_dump(&g_buffer[n], 1);
		jtag_bits_tdiout(g_buffer[n], nbits);
		jtag_show_state();
		// Goto XENDDR state:
		if (mode & F_PAUSE_DR) {
			jtag_goto_state(g_jtag, s_jtag_pause);
		} else {
			jtag_goto_state(g_jtag, s_jtag_shift_frompause);
		}
	} else {
		jtag_bits_tdiout_stay(g_buffer[n], nbits);
	}
	printf("Exit drscan\n");
	jtag_show_state();
	return 0;
}
#endif

int drscan(JtagRegister *tdi, JtagRegister *tdo, short mode, int time)
{
	int error;
	int state;

	// Enter shift DR state:
	jtag_get(g_jtag, JTAG_STATE, &state);
	if (mode & F_BEGIN) {
		if (state == PAUSE_DR) {
			jtag_goto_state(g_jtag, s_jtag_shiftdr_frompause);
		} else if (state == PAUSE_IR) {
			// In this case, walk through RTI
			jtag_goto_state(g_jtag, s_jtag_rtishiftdr_frompause);
		} else {
			jtag_goto_state(g_jtag, s_jtag_shift_dr);
		}
	}

	jtag_get(g_jtag, JTAG_STATE, &state);
	if (state != SHIFT_DR) {
		printf("ERROR: Not in Shift DR state (%s)\n",
				g_statestrings[state]);
		return -1;
	}
	int n = NBYTES(tdi->nbits);
	nextBlock(g_buffer, n);

#ifdef DEBUG
	if (n < 8) {
		printf("     data out:");
		hexdump(g_buffer, n);
	}
#endif

	reverse_buffer(g_buffer, n);
	memcpy(tdi->data, g_buffer, n);

	printf("downloading...\n");
	if (mode & F_END) {
		error = shift_generic(tdi, tdo, tdi->nbits, EXIT);
	} else {
		error = shift_generic(tdi, tdo, tdi->nbits, STAY);
	}

	if (error < 0) {
		printf("Error in shift_generic()\n");
		return error;
	}
	if (tdo) {
		nextBlock(g_buffer, n);
		reverse_buffer(g_buffer, n);
		if (mask_cmp(tdo->data, g_buffer, g_mask, n) != 0) {
			printf("TDO does not match!\n");
			hexdump(tdo->data, (tdo->nbits + 7) / 8);
			printf("Should be:\n");
			hexdump(g_buffer, n);
			return -1;
		}
	}

	if (mode & F_END) {
		// Goto XENDDR state:
		if (mode & F_PAUSE_DR && time == 0) {
			jtag_goto_state(g_jtag, s_jtag_pause);
		} else {
			jtag_goto_state(g_jtag, s_jtag_rti);
			usleep(time);
		}
	}
	return 0;
}

/**
 * Parse and execute XSVF stream
 *
 */

int xsvf_play(JTAG_CTRL jtag)
{
	int error = 0;
	int b;
	unsigned long cur_size = 0;
	unsigned long nmaskbits = 0;
	unsigned long time = 0;
	unsigned long t;
	short flags = 0; // in F_PAUSE_IR flag
	short val;
	short waitstate, endstate;
	int irlen;
	int state;

	JtagRegister *tdi = 0;
	JtagRegister *tdo = 0;

	g_jtag = jtag;

	jtag_get(jtag, JTAG_IRLENGTH, &irlen);

	while (!error) {
		b = nextByte();
		long n = NBYTES(cur_size);

		switch (b) {
		case XCOMPLETE:
			printf("COMPLETE\n");
			return 0;
		case XTDOMASK:
			nextBlock(g_mask, n);
			PRINTF("TDOMASK [%ld]\n", n);
			reverse_buffer(g_mask, n);
			// hexdump(g_mask, n);
			break;
		case XSIR:
			val = nextByte();                  // number of bits
			PRINTF("SIR (%d)\n", val);
			if (val != irlen) {
				printf("IR size in XSVF (%d) does not match chain length (%d).\n"
				       "Aborting.\n", val, irlen);
				return -1;
			}

			jtag_get(g_jtag, JTAG_STATE, &state);
			if (state == PAUSE_IR) {
				error = irscan(val, flags | F_FROM_PAUSE_IR, time);
			} else {
				error = irscan(val, flags, time);
			}
			break;
		case XSDR:
			PRINTF("SDR\n");
			error = drscan(tdi, 0, flags | F_BEGIN | F_END, time);
			break;
		case XRUNTEST:
			time = read_int32();
			PRINTF("RUNTEST %lu\n", time);

			break;
		case XREPEAT:
			val = nextByte();
			PRINTF("REPEAT %d\n", val);
			break;
		case XSDRSIZE:  // set current SDR value size
			PRINTF("SDRSIZE\n");
			cur_size = read_int32();
			if (NBYTES(cur_size) > BLOCKSIZE) {
				return -2;
			}
			if (tdi) {
				reg_free(tdi);
				reg_free(tdo);
			}
			tdi = reg_new(cur_size, JTAGREG_LSB);
			tdo = reg_new(cur_size, JTAGREG_LSB);

			PRINTF("    Size: %lu\n", cur_size);
			break;
		case XSDRTDO:
			PRINTF("SDRTDO [%lu]\n", cur_size);
			error = drscan(tdi, tdo, flags | F_BEGIN | F_END, time);
			break;
		case XSETSDRMASKS:
			PRINTF("SETSDRMASKS ");
			error = set_sdrmasks(cur_size);
			break;
		case XSDRINC:
			PRINTF("XSDRINC\n");
			skip_bytes(n);
			PRINTF(":");
			val = nextByte();
			skip_bytes(val * NBYTES(nmaskbits));
			break;
		case XSDRB:
			PRINTF("SDRB\n");
			// goto shift-DR and shift in data. Stay in shift-DR state
			// after operation.
			error = drscan(tdi, 0, F_BEGIN, time);
			break;
		case XSDRC:
			PRINTF("SDRC\n");
			// shift in data. Stay in shift-DR state after operation.
			error = drscan(tdi, 0, F_CONTINUE, time);
			break;
		case XSDRE:
			PRINTF("SDRE\n");
			// shift in data, and enter XENDDR state when completed.
			error = drscan(tdi, 0, flags | F_END, time);
			break;
		case XSDRTDOB:
			PRINTF("SDRTDOB\n");
			error = drscan(tdi, tdo, F_BEGIN, time);
			break;
		case XSDRTDOC:
			PRINTF("SDRTDOC\n");
			error = drscan(tdi, tdo, F_CONTINUE, time);
			break;
		case XSDRTDOE:
			PRINTF("SDRTDOE\n");
			error = drscan(tdi, tdo, flags | F_END, time);
			break;
		case XSTATE:
			val = nextByte();
			PRINTF("STATE %s\n", g_statestrings[val]);
			error = jtag_moveto_state(g_jtag, val);
			break;
		case XENDIR:
			val = nextByte();
			if (val) {
				flags |= F_PAUSE_IR;
			} else {
				flags &= ~F_PAUSE_IR;
			}
			PRINTF("ENDIR %d\n", val);

			break;
		case XENDDR:
			val = nextByte();
			if (val) {
				flags |= F_PAUSE_DR;
			} else {
				flags &= ~F_PAUSE_DR;
			}
			PRINTF("ENDDR %d\n", val);

			break;

		case XCOMMENT:
			while (nextByte());
			break;
		case XWAIT:
			waitstate = nextByte();
			endstate = nextByte();
			t = read_int32();


			PRINTF("WAIT %lu us in '%s', end in '%s'\n",
				t, g_statestrings[waitstate], g_statestrings[endstate]);

			jtag_moveto_state(g_jtag, waitstate);
			jtag_wait(g_jtag, t);
			error = jtag_moveto_state(g_jtag, endstate);

			break;

#ifdef TEST
		case EOF:
			printf(">> EOF\n");
			return EOF;
#endif
		default:
			printf("Unsupported command: %x\n", b);
			error = -1;
		}
	}
	return error;
}


